/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: opencataloga.c,v 1.24 2018/02/25 21:18:42 piru Exp $

    Desc:
    Lang: english
*/
#define AROS_ALMOST_COMPATIBLE

#include <string.h>
#include <sys/param.h>

#include <hardware/byteswap.h>
#include <libraries/hashtable.h>
#include <libraries/iffparse.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/iffparse.h>
#include <proto/hashtable.h>
#include <proto/utility.h>

#include "locale_intern.h"

#include <aros/debug.h>

#define	DEBUG_OPENCATALOG(x)	;

ULONG ConvertUTF8ToUCS4(CONST_STRPTR src, WSTRPTR dst, LONG maxlen);
ULONG ConvertUCS4ToUTF8(CONST_WSTRPTR src, STRPTR dst, LONG maxlen);

ULONG UTF8_DecodeSafe(CONST_STRPTR utf8, WCHAR *utf32, ULONG length);
ULONG UTF8_Decode(CONST_STRPTR utf8, WCHAR *utf32);
ULONG UTF8_Encode(WCHAR utf32, STRPTR utf8);

#define MIN_CAT_LENGTH 8

/**********************************************************************
	Examine final text length
**********************************************************************/

typedef size_t (*textlenhook)(CONST_APTR src, int srclen);
typedef size_t (*textconvhook)(CONST_APTR src, int srclen, APTR dst);

static size_t textlen_latin1_to_utf8(CONST_APTR _src, int srclen)
{
	UBYTE *src = (UBYTE *)_src;
	ULONG len = 1;

	while (srclen > 0)
	{
		TEXT dummy[4];
		ULONG octets = UTF8_Encode(*src++, (STRPTR)&dummy);

		srclen -= 1;
		len += octets;
	}

	return len;
}

static size_t textlen_latin1_to_utf32(CONST_APTR _src, int srclen)
{
	return srclen * 4 + 4;
}

static size_t textlen_utf32_to_utf8(CONST_APTR _src, int srclen)
{
	CONST_WSTRPTR src = _src;
	ULONG len = 1;

	while (srclen >= 4)
	{
		TEXT dummy[4];
		ULONG octets = UTF8_Encode(*src++, (STRPTR)&dummy);

		srclen -= 4;
		len += octets;
	}

	return len;
}

static size_t textlen_utf8_to_latin1(CONST_APTR _src, int srclen)
{
	CONST_STRPTR src = _src;
	ULONG len = 1;

	while (srclen > 0)
	{
		WCHAR ucs4;
		ULONG octets = UTF8_DecodeSafe(src, &ucs4, srclen);

		if (octets == 0)
			break;

		srclen -= octets;
		src += octets;
		len += 1;
	}

	return len;
}

static size_t textlen_utf8_to_utf32(CONST_APTR _src, int srclen)
{
	CONST_STRPTR src = _src;
	ULONG len = 4;

	while (srclen > 0)
	{
		WCHAR utf32;
		ULONG octets = UTF8_DecodeSafe(src, &utf32, srclen);

		if (octets == 0)
			break;

		srclen -= octets;
		src += octets;
		len += 4;
	}

	return len;
}

#define GET_NEXT_CAT_STRING(_id, _length, _catbuf, _catlen) \
		LONG _id = LE_SWAPLONG((((ULONG *)_catbuf)[0])); \
		LONG _length = LE_SWAPLONG((((ULONG *)_catbuf)[1])); \
		_catbuf += 8; \
		_catlen -= 8

#define TO_NEXT_CAT_STRING(_catbuf, _catlen, _length) \
	if (_length < 0) \
		_length = -_length; \
	_length = (_length + 3) & ~3; \
	_catbuf += _length; \
	_catlen -= _length

/**********************************************************************
	find_dup()
**********************************************************************/

static int find_dup(CONST_STRPTR buffer, ssize_t buflen, CONST_STRPTR cmpbuf, ssize_t cmpbuflen, size_t count)
{
	while (buflen >= MIN_CAT_LENGTH && count > 0)
	{
		LONG strlen = LE_SWAPLONG((((ULONG *)buffer)[1]));

		buffer += 8;
		count--;

		if (strlen == cmpbuflen)
		{
			if (memcmp(buffer, cmpbuf, strlen) == 0)
			{
				LONG id = LE_SWAPLONG((((ULONG *)buffer)[-2]));
				(((ULONG *)cmpbuf)[-1]) = LE_SWAPLONG(-cmpbuflen);	// Pooled string is indicated by negative length
				(((ULONG *)cmpbuf)[0]) = id;	// string is replaced by id to equivalent string
				return 1;
			}
		}

		buflen -= 8;

		TO_NEXT_CAT_STRING(buffer, buflen, strlen);
	}

	return 0;
}

/**********************************************************************
	create_hash()

	This checks for strings for pooling. If string is pooled its length
	is set negative and replacement string id is written to string buffer.
**********************************************************************/

STATIC BOOL create_hash(struct LocaleBase *LocaleBase, struct IntCatalog *cat, CONST_STRPTR buffer, ssize_t buflen, textlenhook exam)
{
	CONST_STRPTR srcbuf = buffer;
	ssize_t srcbuflen = buflen;

	size_t count = 0, unique = 0, savings = 0;
	size_t strbuf = 0;
	int min_id = min_id, max_id = max_id;

	if (buflen >= MIN_CAT_LENGTH)
	{
		LONG id = LE_SWAPLONG((((ULONG *)buffer)[0]));
		min_id = max_id = id;
	}

	while (buflen >= MIN_CAT_LENGTH)
	{
		GET_NEXT_CAT_STRING(id, strlen, buffer, buflen);

		if ((size_t)buflen < (size_t)strlen)
		{
			// Invalid length. Write max unsigned int (convert() will bail out).
			(((ULONG *)buffer)[-1]) = LE_SWAPLONG(0x7fffffff);
			break;
		}

		savings += strlen;

		if (id < min_id)
			min_id = id;
		else if (id > max_id)
			max_id = id;

		// Remaining buffer length must be at least 4 bytes
		if (buflen < 4 || find_dup(srcbuf, srcbuflen, buffer, strlen, count) == 0)
		{
			savings -= strlen;
			unique++;
			strbuf += exam ? exam(buffer, strlen) : strlen + 1;
		}

		count += 1;

		strlen = (strlen + 3) & ~3;
		buffer += strlen;
		buflen -= strlen;
	}

	//NewRawDoFmt("LOCALE: %ld strings where %ld unique saving %ld bytes\n", (APTR)1, NULL, count, unique, savings);

	cat->ic_Strings = AllocVec(strbuf, MEMF_ANY);
	cat->ic_HashTable = CreateHashTableTags(HASHTABLE_Capacity, count, TAG_DONE);

	return (cat->ic_Strings && cat->ic_HashTable) ? TRUE : FALSE;
}

/**********************************************************************
	Convert text
**********************************************************************/

static size_t textconv_noconv(CONST_APTR src, int srclen, APTR dst)
{
	STRPTR dstbuf = dst;
	bcopy(src, dst, srclen);
	dstbuf[srclen] = '\0';
	return srclen + 1;
}

static size_t textconv_utf8_to_latin1(CONST_APTR src, int srclen, APTR dst)
{
	CONST_STRPTR srcbuf = src;
	STRPTR dstbuf = dst;
	ULONG len = 1;

	while (srclen > 0)
	{
		WCHAR ucs4;
		ULONG octets = UTF8_DecodeSafe(srcbuf, &ucs4, srclen);

		if (octets == 0)
			break;

		if (ucs4 > 0xff)
			ucs4 = '?';

		*dstbuf++ = (UBYTE)ucs4;
		srclen -= octets;
		srcbuf += octets;
		len += 1;
	}

	*dstbuf = '\0';

	return len;
}

static size_t textconv_utf8_to_utf32(CONST_APTR src, int srclen, APTR dst)
{
	CONST_STRPTR srcbuf = src;
	WSTRPTR dstbuf = dst;
	ULONG len = 4;

	while (srclen > 0)
	{
		ULONG octets = UTF8_DecodeSafe(srcbuf, dstbuf, srclen);

		if (octets == 0)
			break;

		dstbuf += 1;
		srclen -= octets;
		srcbuf += octets;
		len += 4;
	}

	*dstbuf = '\0';

	return len;
}

static size_t textconv_latin1_to_utf8(CONST_APTR src, int srclen, APTR dst)
{
	CONST_STRPTR srcbuf = src;
	STRPTR dstbuf = dst;
	ULONG len = 1;

	while (srclen > 0)
	{
		ULONG octets = UTF8_Encode(*srcbuf++, dstbuf);	// Returns zero on a NULL

		dstbuf += octets ? octets : 1;
		srclen -= 1;
		len += octets ? octets : 1;
	}

	*dstbuf = '\0';

	return len;
}

static size_t textconv_latin1_to_utf32(CONST_APTR src, int srclen, APTR dst)
{
	CONST_STRPTR srcbuf = src;
	STRPTR dstbuf = dst;
	ULONG len = 4;

	while (srclen > 0)
	{
		*dstbuf++ = *srcbuf++;
		srclen -= 1;
		len += 4;
	}

	*dstbuf = '\0';

	return len;
}

static size_t textconv_utf32_to_utf8(CONST_APTR src, int srclen, APTR dst)
{
	CONST_WSTRPTR srcbuf = src;
	STRPTR dstbuf = dst;
	ULONG len = 1;

	while (srclen > 0)
	{
		ULONG octets = UTF8_Encode(*srcbuf++, dstbuf);	// Returns zero on a NULL

		dstbuf += octets ? octets : 1;
		srclen -= 4;
		len += 1;
	}

	*dstbuf = '\0';

	return len;
}

static void convert(struct LocaleBase *LocaleBase, struct IntCatalog *catalog, STRPTR buffer, int buflen, textconvhook convhook)
{
	APTR hash = catalog->ic_HashTable;
	STRPTR textbuf = catalog->ic_Strings;

	while (buflen >= MIN_CAT_LENGTH)
	{
		STRPTR ptr = textbuf;
		GET_NEXT_CAT_STRING(id, strlen, buffer, buflen);

		if (buflen < strlen)
			break;

		if (strlen < 0)	// String is pooled.
		{
			APTR p;

			GetHashDataByKey(hash, ((ULONG *)buffer)[0], &p);
			ptr = p;
		}
		else
		{
			// Malformed catalog
			if (buflen < strlen)
				break;

			textbuf += convhook(buffer, strlen, textbuf);
		}

		InsertHash(hash, id, ptr);

		TO_NEXT_CAT_STRING(buffer, buflen, strlen);
	}
}

STATIC LONG convert_strings(struct LocaleBase *LocaleBase, struct IntCatalog *catalog, ULONG sourcecodeset, ULONG destcodeset, STRPTR buffer, ssize_t buflen)
{
	LONG rc;
	textlenhook hook = NULL;
	textconvhook convhook = textconv_noconv;

	switch (sourcecodeset)
	{
		case CODESET_LATIN1:
			if (destcodeset == CODESET_UTF8)
			{
				hook = textlen_latin1_to_utf8;
				convhook = textconv_latin1_to_utf8;
			}
			else if (destcodeset == CODESET_UTF32)
			{
				hook = textlen_latin1_to_utf32;
				convhook = textconv_latin1_to_utf32;
			}
			break;

		case CODESET_UTF8:
			if (destcodeset == CODESET_UTF32)
			{
				hook = textlen_utf8_to_utf32;
				convhook = textconv_utf8_to_utf32;
			}
			else if (destcodeset == CODESET_LATIN1)
			{
				hook = textlen_utf8_to_latin1;
				convhook = textconv_utf8_to_latin1;
			}
			break;

		case CODESET_UTF32:
			if (destcodeset == CODESET_UTF8)
			{
				hook = textlen_utf32_to_utf8;
				convhook = textconv_utf32_to_utf8;
			}
			break;
	}

	rc = IFFERR_NOMEM;
	if (create_hash(LocaleBase, catalog, buffer, buflen, hook))
	{
		convert(LocaleBase, catalog, buffer, buflen, convhook);

		rc = 0;
	}

	return rc;
}

static size_t OpenCatIFF(struct LocaleBase *LocaleBase, struct IFFHandle *iff, CONST_STRPTR basedir, CONST_STRPTR langname)
{
	TEXT filename[512];

	strcpy(filename, basedir);
	AddPart(filename, langname, sizeof(filename));

	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: filename <%s>\n",filename));
	iff->iff_Stream = (IPTR)Open(filename, MODE_OLDFILE);
	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: iffstream 0x%lx\n", iff->iff_Stream));
	return iff->iff_Stream != NULL;
}

static struct IntCatalog *InternalOpenCatalog(struct Locale *locale, CONST_STRPTR name, CONST struct TagItem *tags, struct LocaleBase *LocaleBase);
static struct IntCatalog *CreateSecondaryCatalog(struct IntCatalog *cat, struct BuiltInStrings *bis, struct LocaleBase *LocaleBase);

/*****************************************************************************

    NAME */
#include <proto/locale.h>

	AROS_LH3(struct Catalog *, OpenCatalogA,

/*  SYNOPSIS */
	AROS_LHA(struct Locale  *, locale, A0),
	AROS_LHA(STRPTR          , name, A1),
	AROS_LHA(struct TagItem *, tags, A2),

/*  LOCATION */
	struct LocaleBase *, LocaleBase, 25, Locale)

/* ---- locale.library/OpenCatalogA
 *
 *   Name
 *        OpenCatalogA --
 *
 *   FUNCTION
 *        Open a catalog for translated strings.
 *
 *   INPUTS
 *        locale - desired language for the catalog or NULL to use
 *                 user defined default.
 *        name   - catalog name to open
 *        tags   - optional taglist
 *
 *        OC_BuiltInStrings (V52.4 / MorphOS 3.10)
 *           Pointer to built-in strings. When translation is not available
 *           built-in string is used. This allows using NULL string argument
 *           with GetCatalogStr().
 *
 *   RESULT
 *
 *   NOTES
 *
 *   EXAMPLE
 *
 *   BUGS
 *
 *   SEE ALSO
 *
 * ----
 *
 *   INTERNALS
 *
 *   HISTORY
 *   27-11-96   digulla automatically created from
 *              locale_lib.fd and clib/locale_protos.h
 */
/*****************************************************************************/
{
	AROS_LIBFUNC_INIT
	AROS_LIBBASE_EXT_DECL(struct Library *,LocaleBase)

	struct Locale *def_locale = NULL;
	struct BuiltInStrings *bis;
	struct IntCatalog *cat;

	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: locale 0x%lx name <%s> Tags 0x%lx localebase 0x%lx\n", locale, name, tags, LocaleBase));

#ifdef __MORPHOS__
	TZCheckSysTimeZone(LocaleBase);
#endif

	/* Clear error condition before we start. */
	SetIoErr(0);

	if (!locale)
	{
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: no locale, use default locale\n"));
		locale = def_locale = _OpenDefaultLocale(LocaleBase); //OpenLocale(NULL);
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: def_locale 0x%lx\n", def_locale));
	}

	cat = InternalOpenCatalog(locale, name, tags, LocaleBase);
	bis = (APTR)GetTagData(OC_BuiltInStrings, NULL, tags);

	if (bis)
	{
		cat = CreateSecondaryCatalog(cat, bis, LocaleBase);
	}

	/* Close locale if opened by us */
	_InternalCloseLocale(def_locale, LocaleBase);

	return &cat->ic_Catalog;

	AROS_LIBFUNC_EXIT
} /* OpenCatalogA */


/*
 * Catalogs with secondary strings are private. They are copies from public catalogs (if it exists)
 * but have an application specific hash table for application specific built-in strings.
 *
 * They have pointer to the original catalog for reference counting and private catalogs can be
 * identified with ic_Strings == NULL check.
 *
 */
static struct IntCatalog *CreateSecondaryCatalog(struct IntCatalog *cat, struct BuiltInStrings *bis, struct LocaleBase *LocaleBase)
{
	struct IntCatalog *ic = AllocMem(sizeof(*ic), cat ? MEMF_ANY : MEMF_CLEAR);

	if (ic)
	{
		APTR ht, src_ht = NULL;
		int count = 0;
		BOOL success = FALSE;

		if (cat)
		{
			src_ht = cat->ic_HashTable;
			count = GetHashTableAttribute(src_ht, HASHTABLE_Count);
			bcopy(cat, ic, sizeof(*ic));

			ic->ic_Strings = NULL;
			ic->ic_RootCatalog = cat;
		}

		ic->ic_HashTable = ht = CreateHashTableTags(count > 0 ? HASHTABLE_Count : TAG_DONE, count, TAG_DONE);

		if (ht)
		{
			ssize_t size = bis->Size;
			CONST BYTE *ptr = bis->Data;
			success = TRUE;

			switch (bis->Type)
			{
				case TYPE_CATCOMP_ARRAY:
					while (success && size >= 8)
					{
						ULONG *p = (ULONG *)ptr;
						size_t key = p[0];
						APTR val = (APTR)p[1];

						size -= 8;
						ptr += 8;

						if (src_ht)
							GetHashDataByKey(src_ht, key, &val);

						success = InsertHash(ht, key, val);
					}
					break;

				case TYPE_CATCOMP_BLOCK:
					while (success && size >= 6)
					{
						ULONG *p = (ULONG *)ptr;
						UWORD *w = (WORD *)(ptr + 4);
						size_t key = p[0];
						APTR val = (APTR)(ptr + 6);

						if (src_ht)
							GetHashDataByKey(src_ht, key, &val);

						success = InsertHash(ht, key, val);

						size -= 6 + *w;
						ptr += 6 + *w;
					}
					break;
			}
		}

		if (!success)
		{
			DeleteHashTable(ht);
			FreeMem(ic, sizeof(*ic));
			ic = NULL;
		}
	}

	if (!ic)
		SetIoErr(ERROR_NO_FREE_STORE);

	return ic ? ic : cat;
}

static struct IntCatalog *InternalOpenCatalog(struct Locale *locale, CONST_STRPTR name, CONST struct TagItem *tags, struct LocaleBase *LocaleBase)
{
	struct IntCatalog *catalog = NULL;
	char              *language;
	char              *app_language;
	char              *specific_language;
	struct Process    *MyProcess;
#define FILENAMESIZE 256
	char   filename[FILENAMESIZE];
	LONG   chars;
	ULONG  version;
	ULONG  catversion, catrevision;
	WORD   pref_language;
	UWORD  i;
	ULONG  codeset;
	ULONG  systemcatalog = GetTagData(OC_SystemCatalog, FALSE, tags); /* System catalogs are exclusively loaded from MOSSYS: only. */

	MyProcess = (struct Process *) FindTask(NULL);

	specific_language = (char *)GetTagData(OC_Language, (IPTR)0, tags);

	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: specific lang 0x%lx \"%s\"\n",specific_language,specific_language ? specific_language : "<NULL>"));
	if (specific_language)
	{
		language = specific_language;
		pref_language = -1;
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: language 0x%lx \"%s\"\n",language,language));
	}
	else
	{
		language = locale->loc_PrefLanguages[0];
		pref_language = 0;
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: language 0x%lx \"%s\"\n",language,language));
	}

	if (!language)
	{
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: no language.. return NULL\n"));
		return NULL;
	}

	/*
	 * Check whether the built in language of the application matches
	 * the language of the default locale. If it matches, then I
	 * don't need to load anything.
	 */
	app_language = (char *)GetTagData(OC_BuiltInLanguage,
	                                  (IPTR)defLocale.loc_PrefLanguages[0],  /* This struct is const, value equals "english", simply avoiding dupe strings. */
	                                  tags);

	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: app_language 0x%lx \"%s\"\n", app_language, app_language));

	/*
	 * Note: strcmp() here, not Stricmp(). - Piru
	 */
	if (app_language && !strcmp(app_language, language))
	{
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: use built-in language, return NULL\n"));
		return NULL;
	}

	version = GetTagData(OC_Version, 0, tags);
	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: version %ld\n",version));

	codeset = GetTagData(OC_CodeSet, CODESET_LATIN1, tags);

	if (codeset >= CODESET_COUNT)
	{
		codeset = CODESET_LATIN1;
	}

	if (name)
	{
		struct IFFHandle * iff = NULL;
		TEXT langname[FILENAMESIZE];
		APTR oldwinptr;

		/*
		 * The wanted catalog might be in the list of catalogs that are
		 * already loaded. So check that list first.
		 */

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: CatalogLock 0x%lx\n", &IntLB(LocaleBase)->lb_CatalogLock));

		ObtainSemaphore(&IntLB(LocaleBase)->lb_CatalogLock);

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: search cached Catalog, name \"%s\", language \"%s\"\n", name, language))

		ForeachNode(&IntLB(LocaleBase)->lb_CatalogList, (struct Node *)catalog)
		{
			/*
			 * 27-mar-2003 bugfix
			 * Note: Stricmp() here, not strcmp()! This fixed the problems with
			 * franais catalogs, for example. Also added the missing version
			 * check. - Piru
			 */
			if ((catalog->ic_Catalog.cat_CodeSet == codeset && catalog->ic_Name) &&
			    (!Stricmp(catalog->ic_Name, name))                               &&
			    (version == 0 || (catalog->ic_Catalog.cat_Version == version))   &&
			    (catalog->ic_Catalog.cat_Language)                               &&
			    (!Stricmp(catalog->ic_Catalog.cat_Language, language)))
			{
				DEBUG_OPENCATALOG(dprintf("OpenCatalogA: found Catalog 0x%lx ic_Name \"%s\" cat_Language \"%s\"\n", catalog, catalog->ic_Name, catalog->ic_Catalog.cat_Language));
				catalog->ic_UseCount++;
				ReleaseSemaphore(&IntLB(LocaleBase)->lb_CatalogLock);

				SetIoErr(ERROR_ACTION_NOT_KNOWN);

				DEBUG_OPENCATALOG(dprintf("OpenCatalogA: return Catalog 0x%lx\n",catalog));
				return catalog;

			}

		} /* foreachnode */

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: found none\n"));
		ReleaseSemaphore(&IntLB(LocaleBase)->lb_CatalogLock);

		iff = AllocIFF();

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: iff 0x%lx\n",iff));
		if (!iff)
		{
			SetIoErr(ERROR_NO_FREE_STORE);

			DEBUG_OPENCATALOG(dprintf("OpenCatalogA: no iff, return NULL\n"));
			return NULL;
		}

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: pref_language %ld language 0x%lx \"%s\"\n", pref_language, language, language));

		stccpy(langname, language, FILENAMESIZE);
		AddPart(langname, name, FILENAMESIZE);

		oldwinptr = MyProcess->pr_WindowPtr;
		MyProcess->pr_WindowPtr = (APTR)-1;

		while ((pref_language < 10) && language)
		{
			if (app_language && !strcmp(app_language, language))
			{
				MyProcess->pr_WindowPtr = oldwinptr;
				FreeIFF(iff);
				SetIoErr(0);
				return NULL;
			}

			if (!systemcatalog && MyProcess->pr_HomeDir)
			{
				if (OpenCatIFF(LocaleBase, iff, "PROGDIR:Catalogs", langname))
					break;
			}

			if (OpenCatIFF(LocaleBase, iff, "MOSSYS:LOCALE/Catalogs", langname))
				break;

			if (!systemcatalog)
			{
				if (OpenCatIFF(LocaleBase, iff, "LOCALE:Catalogs", langname))
					break;
			}

			pref_language++;
			language = locale->loc_PrefLanguages[pref_language];

		} /* while (pref_language < 10 && language) */

		MyProcess->pr_WindowPtr = oldwinptr;

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: language 0x%lx \"%s\"\n",language,language ? language : "<NULL>"));

	#undef FILENAMESIZE

		if (!iff->iff_Stream)
		{
			DEBUG_OPENCATALOG(dprintf("OpenCatalogA: no stream, return NULL\n"));
			FreeIFF(iff);
			SetIoErr(ERROR_OBJECT_NOT_FOUND);
			return NULL;
		}

		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: catalog name <%s>\n",name));
		catalog = AllocVec(sizeof(struct IntCatalog) + strlen(name) + 1, MEMF_CLEAR);
		DEBUG_OPENCATALOG(dprintf("OpenCatalogA: catalog 0x%lx\n",catalog));
		if (!catalog)
		{
			DEBUG_OPENCATALOG(dprintf("OpenCatalogA: no memory for catalog, return NULL\n"));
			Close((BPTR)iff->iff_Stream);
			FreeIFF(iff);
			SetIoErr(ERROR_NO_FREE_STORE);
			return NULL;
		}

		catalog->ic_UseCount = 1;
		catalog->ic_Catalog.cat_Link.ln_Name = catalog->ic_Name;
		catalog->ic_Catalog.cat_Language = catalog->ic_LanguageName;
		catalog->ic_Catalog.cat_CodeSet = codeset;
		strcpy(catalog->ic_Name, name);

		InitIFFasDOS(iff);

		if (!OpenIFF(iff, IFFF_READ))
		{
			ULONG catcodeset = CODESET_LATIN1;
			ssize_t stringchunklen = 0;
			APTR stringchunk = NULL;

			while (1)
			{
				LONG error = ParseIFF(iff, IFFPARSE_STEP);

				if (error == IFFERR_EOF)
				{
					DEBUG_OPENCATALOG(dprintf("OpenCatalogA: parsed catalog\n"));

					if (stringchunk)
					{
						error = convert_strings(LocaleBase, catalog, catcodeset, codeset, stringchunk, stringchunklen);
						FreeMem(stringchunk, stringchunklen);

						/*  Did everything go fine? 
						 */
						if (!error)
						{
							if (catalog->ic_LanguageName[0] == '\0')
							{
								/*  No ID_LANG chunk found. So setup languagename ourselves.
								 *  Hmm ... maybe this should be done anyway always. Because
								 *  if the catalog file *does* contain an ID_LANG chunk then
								 *  this should be the same as "language" anyway. And if it
								 *  is not, then maybe OpenCatalogA() should fail :-\ 
								 */

								/* We should check if language name set by ID_LANG is correct. It is possible
								 * someone just copied catalog to wrong location.
								 */
								strcpy(catalog->ic_LanguageName, language);
							}

							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: Adding catalog language \"%s\"\n", catalog->ic_Catalog.cat_Language));

							/* Connect this catalog to the list of catalogs 
							 */
							ObtainSemaphore(&IntLB(LocaleBase)->lb_CatalogLock);
							ADDHEAD((struct List *)&IntLB(LocaleBase)->lb_CatalogList, &catalog->ic_Catalog.cat_Link);
							ReleaseSemaphore(&IntLB(LocaleBase)->lb_CatalogLock);

							CloseIFF(iff);
							Close((BPTR)iff->iff_Stream);
							FreeIFF(iff);

							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: success, return catalog 0x%lx\n",catalog));
							return catalog;
						}
					}
				} /* error == OFFERR_EOF */

				if (error == IFFERR_EOC) /* end of chunk */
				{
					continue;
				}

				if (error == 0)
				{
					struct ContextNode * top = CurrentChunk(iff);
					ssize_t size = top->cn_Size;
					char   buf[100];

					switch (top->cn_ID)
					{
						case ID_FORM:
							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FORM\n"));
							break;

						case ID_FVER:
							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER\n"));

							if (size > (ssize_t)sizeof(buf)) 
							{
								break; /*max 100 bytes*/
							}

							error = ReadChunkBytes(iff, buf, size);

							if (error == size)
							{
								error = 0;
							}
							else
							{
								DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER ReadChunkBytes failed\n"));
								break;
							}

							buf[sizeof(buf) - 1] = 0;

							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER verstring \"%s\"\n", buf));

							/*  Ok now we want to get the version and the revision.
							 *  They are separated by a blank space
							 */
							i = 0;

							while (buf[i] && (buf[i] != ' '))
							{
								i++;
							}

							while (buf[i] && (buf[i] == ' '))
							{
								i++;
							}

							while (buf[i] && (buf[i] != ' '))
							{
								i++;
							}

							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER version \"%s\"\n", &buf[i]));

							if (buf[i])
							{
								if ((chars = StrToLong(&buf[i],(LONG *)&catversion)) < 0)
								{
									break;
								}

								DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER catversion %ld\n", catversion));

								i += chars;
								if (buf[i++])
								{
								    if (StrToLong(&buf[i],(LONG *)&catrevision) < 0)
									{
										break;
									}
								}

								DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER catrevision %ld\n", catrevision));

								catalog->ic_Catalog.cat_Version  = catversion;
								catalog->ic_Catalog.cat_Revision = catrevision;

								if (version != 0 && (catversion != version))
								{
									DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_FVER wrong version (%ld != %ld), RETURN_ERROR\n", catversion, version));
									error = RETURN_ERROR;
								}
							}
							/*
							 * 27-mar-2003 bugfix
							 * This break was located inside the above closing bracket.
							 * That caused the routine to drop into ID_LANG if version
							 * wasn't properly parsed. - Piru
							 *
							 * 07-nov-2007 note
							 * With a proper intendation this would never have happened.
							 * Bad bad AROS coding style. - tokai
							 */
							break;

						case ID_LANG:
							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_LANG\n"));

							/* The IntCatalog structure has only 32 bytes reserved for
							 * the language name. So make sure the chunk is not bigger.
							 */

							if (size > (ssize_t)(sizeof(catalog->ic_LanguageName) - 1))
							{
								break;
							}

							error = ReadChunkBytes(iff, catalog->ic_LanguageName, size);
							if (error == size)
							{
								error = 0;
							}
							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_LANG catalog->ic_LanguageName \"%s\"\n", catalog->ic_LanguageName));

							break;

						case ID_CSET:
							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_CSET\n"));

							if (size == sizeof(struct CodeSet))
							{
								if (size == ReadChunkBytes(iff, &catalog->ic_CodeSet, size))
								{
									catcodeset = catalog->ic_CodeSet.cs_CodeSet;

									if (catcodeset >= CODESET_COUNT)
									{
										catcodeset = CODESET_LATIN1;
									}
								}
							}
							break;

						case ID_STRS:
							DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_STRS\n"));

							stringchunklen = size;

							if (!(stringchunk = AllocMem(stringchunklen, MEMF_ANY)))
							{
								DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_STRS AllocMem failed\n"));
								error = ERROR_NO_FREE_STORE;
								SetIoErr(error);
								break;
							}

							error = ReadChunkBytes(iff, stringchunk, stringchunklen);
							if (error == stringchunklen)
							{
								error = 0;
							}
							else
							{
								DEBUG_OPENCATALOG(dprintf("OpenCatalogA: ID_STRS ReadChunkBytes failed\n"));
								break;
							}
							break;

					} /* switch (top->cn_ID) */

				} /* if (error == 0) */

				if (error)
				{
					DEBUG_OPENCATALOG(dprintf("OpenCatalogA: Error occured, dispose catalog...\n"));

					if (stringchunk)
						FreeMem(stringchunk, stringchunklen);

					/*
					 * An error with the file occurred
					 */
					break;
				}

			} /* while (1) */

			CloseIFF(iff);

		} /* if (!OpenIFF(iff, IFFF_READ)) */

		Close((BPTR)iff->iff_Stream);
		FreeIFF(iff);
		dispose_catalog(catalog, LocaleBase);
		FreeVec(catalog);

	} /* if (name) */

	DEBUG_OPENCATALOG(dprintf("OpenCatalogA: catalog not loaded, return NULL\n"));
	return NULL;
}
